/*  Signifo - automate search for words in dictionary files.
    Copyright (C) 1999 Leandro Bernsmller
  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Fondation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>



// user configurations
#include "config.h"



// program name (update Print_version())
#define PROGRAM_NAME "signifo"

// program version
#define PROGRAM_VERSION "1.0"

// strings size
#define STR_SIZE 1000

// dictionary max line size
#define LINE_SIZE 10000



// verbose mode, usefull for debug
//#define VERBOSE_MODE



// headers
void Print_instructions ();
int Def_parameter_type ( char* );
int Search_word ( char*, char* );
void Print_version ();
int Compare_words ( char*, char* );
void Cut_parenthesis_and_contents ( char*, int );
void Cut_parenthesis ( char*, int );



/*
   Parse command line parameters, open files and start
	dictionary search.
	return one exit code (see program help)
*/
int
main ( int nArg_c, char **pszArg_v )
{
  // verify if receive some command line parameters
  if ( nArg_c <= 1 )
    {
      fprintf ( stderr, "%s: execute with --help option for details\n", PROGRAM_NAME );
      exit ( 1 );
    }

  // init general data

  // the dictionary file name
  char szDict_file [ STR_SIZE ] = DEFAULT_DICT;

  // the word to search for
  char szWord [ STR_SIZE ];      
  szWord [ 0 ] = 0;

  // parse commandline parameters

  // state machine
  int nState = 0;              // control parameters interpretation
  int nCode = 0;               // parameter code
  int ncP = 0;                 // parameters conter
  char szParam [ STR_SIZE ];   // parameter string
  do
    {
      // copy parameter to buffer
      strcpy ( szParam, pszArg_v [ ncP ] );

      // define parameter type - make more or less lexical test
      nCode = Def_parameter_type ( szParam );
      
      // verify if the parameter is flaging to use stdin as input
      if ( nCode == 4 )
        {
          // replace parameter string with stdin data
          fgets ( szParam, STR_SIZE - 1, stdin );
          // reformat
          szParam [ strlen ( szParam ) - 1 ] = 0;
          
          // test parameter again
          nCode = Def_parameter_type ( szParam );
        }
      
      // parameters test
      switch ( nState )
        {
          case 0:
            // ignore first parameter (executable name)
            nState = 1;
            break;

          // interpret flags
          case 1:
          
            switch ( nCode )
              {
              case 1:
                nState = 2;
                break;
                
              case 2:
                nState = 3;
                break;
                
              case 3:
                Print_instructions ();
                exit ( 0 );
                break;
                
              case 4:
                // if is the executable name (first parameter)
                if ( ncP == 0 )
                   nState == 1;
                else
                  // invalid parameter
                  {
                    fprintf ( stderr, "%s: %s is not a valid parameter, execute signifo --help for details\n", PROGRAM_NAME, szParam );
                    exit ( 1 );
                  }
                break;
                
              case 5:
                Print_version ();
                exit ( 0 );
                break;
              }
            break;
            
          // receive dictionary file name
          case 2:        
            strcpy ( szDict_file, szParam );
            nState = 1;
            break;
            
          // receive word
          case 3:        
            strcpy ( szWord, szParam );
            nState = 1;
            break;
        }

      // next parameter
      ncP ++;
    }
  while ( ncP < nArg_c );
  
  // try open input file
  FILE
    *pstDict_file = fopen ( szDict_file, "rt" );
  if ( pstDict_file == NULL )
    {
      perror ( PROGRAM_NAME );
      exit ( 1 );
    }

  // exit code if word not found
  int nExit_code = 2;
     
  // make linear search in file
  // read each line
  char szLine [ LINE_SIZE ];       // store one line
  int nFound;                      // number of words found
  while ( ! feof ( pstDict_file ) )
    {
      // read one line
      char* pchRead = fgets ( szLine, STR_SIZE - 1, pstDict_file );
      
      // if error, don't continue
      if ( pchRead == NULL )
        continue;
      
      #ifdef VERBOSE_MODE
      printf ( "\nline %s    ",szLine );
      #endif
      
      // search word in line
      nFound = Search_word ( szLine, szWord );
      
      // if found, output the line
      for ( int ncF = 0; ncF < nFound; ncF ++ )
        {
          fprintf ( stdout, "%s", szLine );
          // at least one word found
          nExit_code = 0;
        }

      #ifdef VERBOSE_MODE
      printf ( " %d ", nFound );
      #endif
    }

  fclose ( pstDict_file );
  
  exit ( nExit_code );
}



/*
   Print program instructions
 */
void
Print_instructions ()
{
  printf ( "Usage: %s [OPTION]... [-d directionary-file] [-s word-to-search]\n", PROGRAM_NAME );
  printf ( "Search for a given word in a dictionary, try output the word meaning to stdout.\n" );
  printf ( "\n" );
  printf ( "  -d FILE        dictionary FILE to consult\n" );
  printf ( "                 (if not defined is used '%s' as default)\n", DEFAULT_DICT );
  printf ( "\n" );
  printf ( "  -s WORD        word to search for\n" );
  printf ( "\n" );
  printf ( "  --help         program usage instructions\n" );
  printf ( "\n" );
  printf ( "  --version      program version and copyright note\n" );
  printf ( "\n" );
  printf ( "  -              can be used to replace one token (flag,\n" );
  printf ( "                 file name or word) with stdin contents\n" );
  printf ( "\n" );
  printf ( "Exit codes:\n" );
  printf ( "  0 - ok, word found\n" );
  printf ( "  1 - general error (file not found,...), see stderr output\n" );
  printf ( "  2 - error, word not found\n" );
  printf ( "\n" );
  printf ( "Files:\n" );
  printf ( "  %s - default dictionary file, used if -d option was not found.\n", DEFAULT_DICT );
  printf ( "\n" );
  printf ( "Bugs (see file TODO):\n" );
  printf ( "  strings and dictionary lines have a limited static size\n" );
  printf ( "\n" );
  printf ( "More information is available in documentation.\n" );
  printf ( "\n" );
  printf ( "Bug reports: lebe@geocities.com\n" );
}



/*
   Print program version
 */
void
Print_version ()
{
  printf ( "Signifo version %s, Copyright (C) 1999 Leandro Bernsmuller\n", PROGRAM_VERSION );
  printf ( "Signifo comes with ABSOLUTELY NO WARRANTY.\n" );
  printf ( "This is free software, and you are welcome to redistribute it\n" );
  printf ( "under certain conditions; see file COPYING for details.\n" );
}



/*
   Make some lexical tests, define the parameter type.
   return one parameter code
 */
int
Def_parameter_type ( char* pszParam )
{
  // the parameter is dictionary file flag
  if ( strcmp ( pszParam, "-d" ) == 0 )
    return 1;
  else
    // the parameter is word flag
    if ( strcmp ( pszParam, "-s" ) == 0 )
      return 2;
    else
      // the parameter is help flag
      if ( strcmp ( pszParam, "--help" ) == 0 )
        return 3;
      else
        // the parameter flags stdin input
        if ( strcmp ( pszParam, "-" ) == 0 )
          return 4;
        else
          // the parameter flags stdin input
          if ( strcmp ( pszParam, "--version" ) == 0 )
            return 5;

  return 0;
}



/*
   Make lexical analisis on the line, searching for the word.
   pszLine - line where make search
   pszWord - word to search for
  
   return
     number of times the word was found in line
 */
int
Search_word ( char* pszLine, char* pszWord )
{
  char szBase_word [ STR_SIZE ];   // base word (in 'vorto', the base is 'vort')
  char szTest_word [ STR_SIZE ];   // one dictionary word to test
  
  // reset buffers
  memset ( szBase_word, 0, STR_SIZE );
  memset ( szTest_word, 0, STR_SIZE );

  // line string size
  int  nSize = strlen ( pszLine );

  // lexical analisis
  
  // current state
  int nState = 1;
  
  // flag word found
  int nFound = 0;

  // scan line
  for ( int ncCar = 0; ncCar < nSize; ncCar ++ )
    {
      // test the caracter in state machine
      switch ( nState )
        {
          // end of line
          case 0:
            return 0;
            break;
            
          // scan for first word
          case 1:
            if ( pszLine [ ncCar ] == '#' )
              return 0;
            else
              if ( pszLine [ ncCar ] == '<' )
                nState = 2;
            break;
            
          // scan base word
          case 2:
            if ( pszLine [ ncCar ] == '/' )
              nState = 3;
            else
              if ( pszLine [ ncCar ] == '>' )
                {
                  // compare
                  int nResult = Compare_words ( pszWord, szTest_word );

                  // word found
                  if ( nResult == 1 )
                    nFound ++;
                
                  // continue line scan
                  nState = 4;
                }
              else
                // add char to base and test word, if - ignore it
                if ( pszLine [ ncCar ] != '-' )
                  {
                    szBase_word [ strlen ( szBase_word ) ] = pszLine [ ncCar ];
                    szTest_word [ strlen ( szTest_word ) ] = pszLine [ ncCar ];
                  }
            break;
                
          // add to test word the base word sufix
          case 3:
            if ( pszLine [ ncCar ] == '>' )
              {
                // compare
                int nResult = Compare_words ( pszWord, szTest_word );

                // word found
                if ( nResult == 1 )
                  nFound ++;
                
                // continue line scan
                nState = 4;
              }
            else
              // add char to test word, if - ignore it
              if ( pszLine [ ncCar ] != '-' )
                szTest_word [ strlen ( szTest_word ) ] = pszLine [ ncCar ];
            break;
                
          // scan for other test words
          case 4:
            if ( pszLine [ ncCar ] == '#' )
              return 0;
            else
              if ( pszLine [ ncCar ] == '<' )
                {
                  nState = 5;

                  // empty test word
                  memset ( szTest_word, 0, STR_SIZE );
                }
            break;
            
          // found other test word in line, add prefix to test word
          case 5:
            if ( pszLine [ ncCar ] == '~' )
              // add to test word the base one
              strcat ( szTest_word, szBase_word );
            else
              if ( pszLine [ ncCar ] == '>' )
                {
                  // compare words
                  int nResult = Compare_words ( pszWord, szTest_word );

                  // word found
                  if ( nResult == 1 )
                    nFound ++;
                
                  // continue line scan
                  nState = 4;
                }
              else
                // add char to test word, if - ignore it
                if ( pszLine [ ncCar ] != '-' )
                  szTest_word [ strlen ( szTest_word ) ] = pszLine [ ncCar ];
            break;
        }
    }
    
  // return words found
  return nFound;
}
                
              
                
/*
   Compare two words.
   pszWord1, pszWord2 - words to compare
   Comparison is case sensitive.
   () can be used to define optional word parts on word 2.
    Example: "a(bc)d" represent "abcd" and "ad"

   return: 0 - not equal / 1 - equal
 */
int
Compare_words ( char* pszWord1, char* pszWord2 )
{
  #ifdef VERBOSE_MODE
  printf ( "  %s == %s?  ", pszWord1, pszWord2 );
  #endif

  // search for (), if found generate alternative words
  // scan word 2
  for ( int ncW2 = 0; ncW2 < strlen ( pszWord2 ); ncW2 ++ )
    if ( pszWord2 [ ncW2 ] == '(' )
      {
        // generate alternative words: with and without parenthesis contents

        // cut () and it's contents
        char szAlter_word2 [ STR_SIZE ];
        memcpy ( szAlter_word2, pszWord2, STR_SIZE );
        Cut_parenthesis_and_contents ( szAlter_word2, ncW2 );
        
        // make comparision with the new word
        int nEqual = Compare_words ( pszWord1, szAlter_word2 );

        // if equal
        if ( nEqual )
          return nEqual;
        else
          {
            // try with the other alternative word
            // cut () only
            memcpy ( szAlter_word2, pszWord2, STR_SIZE );
            Cut_parenthesis ( szAlter_word2, ncW2 );
        
            // make comparison with the new word
            return Compare_words (  pszWord1, szAlter_word2 );
          }
      }

  // compare lenghts
  if ( strlen ( pszWord1 ) != strlen ( pszWord2 ) )
    return 0;

  // check word 0 - not equal / 1 - equal
  char chEqual = 1;

  // scan words
  int ncCar1 = 0;
  int ncCar2 = 0;
  
  // while words have chars and are equal
  while (    chEqual == 1
          && pszWord1 [ ncCar1 ] != 0
          && pszWord2 [ ncCar2 ] != 0 )
   {
     // compare chars
     chEqual = pszWord1 [ ncCar1 ] == pszWord2 [ ncCar2 ];

     // next chars
     ncCar1 ++;
     ncCar2 ++;
   }     
       
  return chEqual;
}


/*
   Remove the first parenthesis pair found and ALL simbols
   between them (including other parenthesis).
   Example: ab(cd)e -> abe
   pszWord - the word,
   nStart  - position for the open parenthesis

   return pszWord - the word with no (..,)
 */
void
Cut_parenthesis_and_contents ( char* pszWord, int nStart )
{
  // scan word, searching for the close parenthesis
  int nEnd = nStart;   // the ) position
  int ncW = nStart;
  while ( nEnd == nStart && pszWord [ ncW ] != 0 )
    {
      if ( pszWord [ ncW ] == ')' )
        nEnd = ncW;
      
      ncW ++;
    }
    
  // copy the simbols after the close parenthesis over the parenthesis area
  int nWord_lenght = strlen ( pszWord );
  int nPare_lenght = nEnd - nStart + 1;
  for ( int ncS = nStart; ncS < nWord_lenght - nPare_lenght + 1; ncS ++ )
    if ( ncS <= nWord_lenght )
      pszWord [ ncS ] = pszWord [ ncS + nEnd - nStart + 1 ];
}



/*
   Remove the first parenthesis pair (mantain the contents).
   Example: ab(cd)e -> abcde
   pszWord - the word,
   nStart  - position for the open parenthesis
   
   return pszWord - the word with no ()
 */
void
Cut_parenthesis ( char* pszWord, int nStart )
{
  // Scan word, searching for the close parenthesis,
  // write simbols over the open parenthesis.
  int ncW = nStart;
  int nNext_char = 1;
  while ( pszWord [ ncW + nNext_char ] != 0 )
    {
      if ( pszWord [ ncW + nNext_char ] == ')' && nNext_char < 2 )
        nNext_char ++;
      pszWord [ ncW ] = pszWord [ ncW + nNext_char ];
      
      ncW ++;
    }

  pszWord [ ncW ] = 0;
}
